#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <stdint.h>
#include <ctype.h>
#include "image.h"
#include "color_bits.h"

uint32_t *create_refvec_map(int xsz, int ysz, uint32_t *nmap, int nx, int ny, float ior);
uint32_t *create_offset_map(int xsz, int ysz, float dist, uint32_t *rvmap);

int main(int argc, char **argv) {
	uint32_t *norm_map, *rvmap = 0, *omap = 0;
	unsigned long nx, ny;
	float dist = 1.0;
	int i, xsz, ysz = -1;
	char *norm_fname = 0;
	char *oref_fname = 0;
	char *out_fname = 0;
	float ior = 1.4;

	for(i=1; i<argc; i++) {
		if(argv[i][0] == '-' && argv[i][2] == 0) {
			switch(argv[i][1]) {
			case 'd':	/* distance */
				if(isdigit(argv[++i][0])) {
					dist = atof(argv[i]);
				} else {
					fprintf(stderr, "-d must be followed by a number (distance)\n");
					return EXIT_FAILURE;
				}
				break;

			case 'i':
				if(isdigit(argv[++i][0])) {
					ior = atof(argv[i]);
				} else {
					fprintf(stderr, "-i must be followed by a number (IOR)\n");
					return EXIT_FAILURE;
				}
				break;

			case 's':	/* output image size */
				{
					char *sep = strchr(argv[++i], 'x');
					if(isdigit(argv[i][0]) && sep && isdigit(*(sep + 1))) {
						xsz = atoi(argv[i]);
						ysz = atoi(sep + 1);
					} else {
						fprintf(stderr, "-s must be followed by a resolution (WxH)\n");
						return EXIT_FAILURE;
					}
				}
				break;
				
			case 'n':	/* normal map */
				norm_fname = argv[++i];
				break;

			case 'r':	/* output refraction dir map */
				oref_fname = argv[++i];
				break;

			case 'o':	/* output offset map */
				out_fname = argv[++i];
				break;
			}
		}
	}

	if(!norm_fname) {
		fprintf(stderr, "you must specify the normal map to use (-n)\n");
		return EXIT_FAILURE;
	}

	if(!out_fname && !oref_fname) {
		fprintf(stderr, "you must specify at least one of -o or -r\n");
		return EXIT_FAILURE;
	}

	if(ysz == -1) {
		printf("didn't specify output size, using default 256x256\n");
		xsz = ysz = 256;
	}

	if(!(norm_map = load_image(norm_fname, &nx, &ny))) {
		perror("failed to open normal map");
		return EXIT_FAILURE;
	}

	printf("creating the reflection vector map...\n");
	if(!(rvmap = create_refvec_map(xsz, ysz, norm_map, nx, ny, ior))) {
		fprintf(stderr, "creation of the reflection vector map, failed\n");
		return EXIT_FAILURE;
	}
	
	printf("saving...\n");
	if(oref_fname) {
		save_image(oref_fname, rvmap, xsz, ysz, IMG_FMT_TGA);
	}

	if(out_fname) {
		printf("creating the hit offset map...\n");
		if(!(omap = create_offset_map(xsz, ysz, dist, rvmap))) {
			fprintf(stderr, "creation of the offset map failed\n");
			return EXIT_FAILURE;
		}
		printf("saving...\n");
		save_image(out_fname, omap, xsz, ysz, IMG_FMT_TGA);
	}
	
	free(rvmap);
	free(omap);
	free_image(norm_map);

	return 0;
}

#define SQ(x)	((x) * (x))

float dot(float x1, float y1, float z1, float x2, float y2, float z2) {
	return x1 * x2 + y1 * y2 + z1 * z2;
}

void reflect(float *rx, float *ry, float *rz, float x, float y, float z, float nx, float ny, float nz) {
	float d = dot(x, y, z, nx, ny, nz);
	*rx = -(2.0 * d * nx - x);
	*ry = -(2.0 * d * ny - y);
	*rz = -(2.0 * d * nz - z);
}

void refract(float *rx, float *ry, float *rz, float x, float y, float z, float nx, float ny, float nz, float sior, float dior) {
	float beta;
	float cos_inc = x * nx + y * ny + z * nz;
	float ior = sior / dior;
	float radical = 1.0 + SQ(ior) * (SQ(cos_inc) - 1.0);

	if(radical < 0.0) {	/* total internal reflection */
		reflect(rx, ry, rz, x, y, z, nx, ny, nz);
		return;
	}

	beta = ior * cos_inc - sqrt(radical);

	*rx = x * ior + nx * beta;
	*ry = y * ior + ny * beta;
	*rz = z * ior + nz * beta;
}

void normalize(float *x, float *y, float *z) {
	float len = (float)sqrt(SQ(*x) + SQ(*y) + SQ(*z));
	*x /= len;
	*y /= len;
	*z /= len;
}


uint32_t pack_vec(float x, float y, float z) {
	return ((((int)(x * 128.0) + 128) << RED_SHIFT32) & RED_MASK32) |
			((((int)(y * 128.0) + 128) << GREEN_SHIFT32) & GREEN_MASK32) | 
			((((int)(z * 128.0) + 128) << BLUE_SHIFT32) & BLUE_MASK32);
}

void unpack_vec(uint32_t pixel, float *x, float *y, float *z) {
	*x = ((pixel & RED_MASK32) >> RED_SHIFT32) / 128.0f - 1.0f;
	*y = ((pixel & GREEN_MASK32) >> GREEN_SHIFT32) / 128.0f - 1.0f;
	*z = ((pixel & BLUE_MASK32) >> BLUE_SHIFT32) / 128.0f - 1.0f;
}

uint32_t *create_refvec_map(int xsz, int ysz, uint32_t *nmap, int nx, int ny, float ior) {
	uint32_t *img, *ptr;
	int i, j;
	
	if(!(img = malloc(xsz * ysz * sizeof(uint32_t)))) {
		return 0;
	}

	ptr = img;

	for(j=0; j<ysz; j++) {
		for(i=0; i<xsz; i++) {
			float normx, normy, normz;
			float rx, ry, rz;
			
			int x = (float)i * (float)nx / (float)xsz;
			int y = (float)j * (float)ny / (float)ysz;

			unpack_vec(nmap[y * nx + x], &normx, &normy, &normz);
			
			refract(&rx, &ry, &rz, 0, 0, 1, normx, normy, normz, 1.0, ior);
			normalize(&rx, &ry, &rz);
			
			*ptr++ = pack_vec(rx, ry, rz);
		}
	}

	return img;
}

uint32_t *create_offset_map(int xsz, int ysz, float dist, uint32_t *rvmap) {
	uint32_t *img, *ptr;
	int i;
	int sz = xsz * ysz;

	if(!(img = malloc(xsz * ysz * sizeof(uint32_t)))) {
		return 0;
	}

	ptr = img;
	for(i=0; i<sz; i++) {
		float rx, ry, rz;
		float xrx, xry, xrz;
		float yrx, yry, yrz;
		float dx, dy;
		float dotx, doty;

		unpack_vec(*rvmap++, &rx, &ry, &rz);

		xrx = rx; xry = 0; xrz = rz;
		yrx = 0; yry = ry; yrz = rz;

		normalize(&xrx, &xry, &xrz);
		normalize(&yrx, &yry, &yrz);

		dotx = dot(xrx, xry, xrz, 0, 0, 1);
		doty = dot(yrx, yry, yrz, 0, 0, 1);

		dx = dist * tan(acos(dotx));
		dy = dist * tan(acos(doty));

		*ptr++ = pack_vec(rx > 0 ? dx : -dx, ry > 0 ? dy : -dy, 0);
	}
	
	return img;
}
